# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

"""Test helpers."""

import StringIO
import icemac.callonchange.observer
import os
import os.path
import shutil
import sys
import tempfile
import unittest


def grapStdout(callable, *args, **kw):
    """Grab the standard output and return it together with the result."""
    orig_stdout = sys.stdout
    sys.stdout = StringIO.StringIO()
    try:
        try:
            result = callable(*args, **kw)
        except SystemExit:
            # Handled internally
            result = ''
        #noinspection PyUnresolvedReferences
        return sys.stdout.getvalue(), result
    finally:
        sys.stdout = orig_stdout


class ObserverTestBase(unittest.TestCase):
    # Base class for directory an file observers

    def setUp(self):
        # Set up a place for the tests.
        self.tempdir = tempfile.mkdtemp()
        # Create the directory the observation normally takes place.
        self.basedir = os.path.join(self.tempdir, 'basedir')
        os.mkdir(self.basedir)

    def tearDown(self):
        # Remove all temporary test directories.
        shutil.rmtree(self.tempdir)

    def createScript(self):
        filename = os.path.join(self.tempdir, 'script')
        file = open(filename, 'w')
        file.write('#! /bin/bash\n')
        file.write('echo -n "script called" > "%s"\n' % os.path.join(
            self.tempdir, 'result'))
        file.flush()
        os.fsync(file.fileno())
        file.close()
        os.chmod(filename, 0o700)
        return filename

    def createObserver(self, dir=None, script=None, quite=True, **kw):
        # Create the observer which calls a predefined script.
        if dir is None:
            # By default observer the default directory.
            dir = self.basedir
        if script is None:
            # By default create an observation script.
            script = self.createScript()
        # Create the observer but make sure it does not exit on error
        # as this will exit the test runner, too.
        return icemac.callonchange.observer.Observer(
            dir, [script], quite=quite, **kw)

    def assertScriptCalled(self):
        """Assert that the script got called."""
        # Wait a bit as the events are processed in a different thread.
        self.assertEqual('script called',
                         open(os.path.join(self.tempdir, 'result')).read())

    def assertScriptNotCalled(self):
        """Assert that the script got not called."""
        # Wait a bit as so the event can get propagated.
        self.assertFalse(os.path.exists(os.path.join(self.tempdir, 'result')))



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

import icemac.callonchange.observer
import icemac.callonchange.testing
import unittest


# Options which are set by default by the mangle function when no
# options are set. After adding new options enter them here:
DEFAULT_OPTIONS = {'extensions': [], 'quite': False, 'immediate': False}


def expected_options(**options):
    """Expected options which differ from default."""
    # Expected options are the default ones updated by the ones
    # different:
    expected = DEFAULT_OPTIONS.copy()
    expected.update(options)
    return expected


class TestMangle(unittest.TestCase):

    def callFUT(self, arg, argv=None):
        if not argv:
            argv = []
        return icemac.callonchange.testing.grapStdout(
            icemac.callonchange.observer.mangle_call_args, arg, argv)

    def test_no_args(self):
        # With no arguments supplied, usage is shown.
        stdout, result = self.callFUT([])
        self.assertEqual((None, None, None), result)
        self.assertTrue(stdout.lower().startswith('usage:'))

    def test_missing_params(self):
        # With not enough arguments supplied, usage is shown.
        stdout, result = self.callFUT(['.'])
        self.assertEqual((None, None, None), result)
        self.assertTrue(stdout.lower().startswith('usage:'))

    def test_only_extension(self):
        # With only extensions supplied, usage is shown.
        stdout, result = self.callFUT(['-e', '.py'])
        self.assertEqual((None, None, None), result)
        self.assertTrue(stdout.lower().startswith('usage:'))

    def test_no_additional_args(self):
        # Without additional arguments the default arguments are used.
        stdout, result = self.callFUT(['.', 'bin/test'])
        self.assertEqual(('.', ['bin/test'], expected_options()), result)
        self.assertEqual(stdout, '')

    def test_additional_args(self):
        # With additional arguments these are added to the default ones.
        stdout, result = self.callFUT(
            ['.', 'bin/test'], ['-t', 'TestMangle'])
        self.assertEqual(('.',
                          ['bin/test', '-t', 'TestMangle'],
                          expected_options()), result)
        self.assertEqual(stdout, '')

    def test_only_additional_args(self):
        # With only additional arguments supplied, they are used
        # instead of the default ones.
        stdout, result = self.callFUT([], ['.', 'bin/test', '-v'])
        self.assertEqual(('.', ['bin/test', '-v'], expected_options()), result)
        self.assertEqual(stdout, '')

    def test_one_extension(self):
        # With an extension supplied it is returned in the third
        # parameter of the return value.
        stdout, result = self.callFUT(['-e', '.py', '.', 'bin/test'])
        self.assertEqual(('.',
                          ['bin/test'],
                          expected_options(extensions=['.py'])), result)
        self.assertEqual(stdout, '')

    def test_two_extensions(self):
        # With more than one extension supplied they are returned in
        # the third parameter of the return value.
        stdout, result = self.callFUT(
            ['-e', '.py', '-e', '.txt', '.', 'bin/test'])
        self.assertEqual(('.',
                          ['bin/test'],
                          expected_options(extensions=['.py', '.txt'])),
                         result)
        self.assertEqual(stdout, '')

    def test_wrong_order_of_extension_an_parameter(self):
        # The (optional) extensions must come before the positional
        # arguments of path and callable.
        stdout, result = self.callFUT(
            ['.', '-e', '.py', 'bin/test'])
        # Option specification becomes part of the callable.
        self.assertEqual(('.', ['-e', '.py', 'bin/test'], expected_options()),
                         result)
        self.assertEqual(stdout, '')

    def test_extension_in_additional_params_not_accepted(self):
        # The extensions in the additional parameters are used as
        # options of the callable not ay arguments of
        # icemac.callonchange.
        stdout, result = self.callFUT(
            ['.', 'bin/test'], ['-e', '.py'])
        # Option specification becomes part of the callable.
        self.assertEqual(('.', ['bin/test', '-e', '.py'], expected_options()),
                         result)
        self.assertEqual(stdout, '')

    def test_extension_without_dot_gets_dot_added(self):
        # When the extension is specified without the leading dot it
        # gets added.
        stdout, result = self.callFUT(
            ['-e', 'py', '.', 'bin/test'])
        # In the result the extensions list contains the extension
        # including a leading dot. This is for convinience.
        self.assertEqual(('.',
                          ['bin/test'],
                          expected_options(extensions=['.py'])), result)
        self.assertEqual(stdout, '')

    def test_quite(self):
        # The option -q makes callonchange quite.
        stdout, result = self.callFUT(['-q', '.', 'bin/test'])
        # The result the options dict contains a key for "quite":
        self.assertEqual(('.',
                          ['bin/test'],
                          expected_options(quite=True)), result)
        self.assertEqual(stdout, '')

    def test_immediate(self):
        # The option -i makes callonchange running immediately after start.
        stdout, result = self.callFUT(['-i', '.', 'bin/test'])
        # The result the options dict contains a key for "immediate":
        self.assertEqual(('.',
                          ['bin/test'],
                          expected_options(immediate=True)), result)
        self.assertEqual(stdout, '')



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

import icemac.callonchange.testing
import os
import os.path
import time


class TestFileObserver(icemac.callonchange.testing.ObserverTestBase):
    # Tests for the observation of specific files in a directory.

    def writeFile(self, path, mode='w'):
        my_file = file(path, mode)
        # my_file.flush and os.fsync are not enough (at least on my
        # MacBook Pro), so I write really big files hopefully beyond
        # the cache size.
        my_file.write(10000000 * 'asdf')
        my_file.flush()
        os.fsync(my_file.fileno())
        my_file.close()

    def callFUT(self, filename, filemode='w', dir=None, script=None,
                quite=True, immediate=False, **kw):
        # Call the function under test.
        # filename ... name of the file which is to be created.
        # filemode ... mode of the created file, defaults to 'w'
        # dir ... directory which is to be observed, defaults to self.basedir
        # script ... name of teh script to be called, defaults to creation of
        #            default script
        # quite ... set quite option on observer
        # **kw ... additional options for observer
        observer = self.createObserver(
            dir=dir, script=script, quite=quite, immediate=immediate, **kw)
        try:
            try:
                observer.start()
                # Wait a bit so the observer can set up its observation
                # routine:
                time.sleep(1)
                if not immediate:
                    self.writeFile(os.path.join(self.basedir, filename),
                                   mode=filemode)
            except KeyboardInterrupt:
                # Do not care about KeyboardInterrupt here as it is raised
                # when the script cannot be found.
                pass
        finally:
            observer.stop()

    def test_not_existing_script(self):
        # When the script does not exist, an error message is
        # displayed.
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, 'one.py', script='file-not-existing-script.sh',
            extensions=['.py'])
        self.assertEqual(
            "OSError: [Errno 2] No such file or directory\n"
            "Parameters were: file-not-existing-script.sh\n", stdout)

    def test_not_existing_script_not_quite(self):
        # When the script does not exist and the quite option is not
        # set, the called script name and an error message are
        # displayed.
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, 'one.py', script='file-not-existing-script.sh',
            quite=False, extensions=['.py'])
        self.assertEqual(
            "Calling: file-not-existing-script.sh\n"
            "OSError: [Errno 2] No such file or directory\n", stdout)

    def test_not_matching_extension(self):
        # When the extension of the changed file inside the observed
        # directory is not in the extenion list, script is not called
        self.callFUT('one.pyc', extensions=['.py'])
        self.assertScriptNotCalled()

    def test_matching_extension_create(self):
        # The script is called when the extension of the created file
        # is inside the observed directory and the file's extension is
        # in the extension list.
        self.callFUT('one.py', extensions=['.py'])
        self.assertScriptCalled()

    def test_matching_extension_modify(self):
        # The script is also called when the extension of the changed
        # file inside the observed directory is in the extenion list.
        # Create a file which gets modified when under observation.
        path = os.path.join(self.basedir, 'one.py')
        self.writeFile(path)
        self.callFUT('one.py', filemode='w+a', extensions=['.py'])
        self.assertScriptCalled()

    def test_matching_extension_not_quite(self):
        # When the quite option is not set, a message about the called
        # script is displayed.
        stdout, result = icemac.callonchange.testing.grapStdout(
            self.callFUT, 'one.py', quite=False, extensions=['.py'])
        self.assertScriptCalled()
        self.assert_(stdout.startswith('Calling: '))
        self.assert_(stdout.endswith('script\n'))

    def test_matching_extension_but_outside_observed_dir(self):
        # When a file outside the observed dir changes the script does
        # not get called.
        observe_dir = os.path.join(self.basedir, 'observe')
        os.mkdir(observe_dir)
        self.callFUT('one.py', dir=observe_dir, extensions=['.py'])
        self.assertScriptNotCalled()

    def test_immediate(self):
        # When "immediate" is set the utility is called without a change.
        self.callFUT('one.py', extensions=['.py'], immediate=True)
        self.assertScriptCalled()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

import icemac.callonchange.testing
import os
import os.path
import time


class TestDirectoryObserver(icemac.callonchange.testing.ObserverTestBase):
    # Tests for the observation of all files in a directory.

    def callFUT(
        self, dir=None, script=None, quite=True, immediate=False, **kw):
        # Call the function under test
        observer = self.createObserver(
            dir=dir, script=script, quite=quite, immediate=immediate, **kw)
        try:
            try:
                observer.start()
                if not immediate:
                    os.mkdir(os.path.join(self.basedir, '1'))
                time.sleep(1)
            except KeyboardInterrupt:
                # When an OSError occurres during script call a
                # KeyboardInterrupt is raised to stop the whole
                # process. But this is the test which should not get
                # stopped.
                pass
        finally:
            observer.stop()

    def test_observer_called_on_changed_dir(self):
        # Assert that the script is called when something inside the
        # observed directory changes.
        self.callFUT()
        self.assertScriptCalled()

    def test_observer_called_on_changed_dir_not_quite(self):
        # Assert that the script is called when something inside the
        # observed directory changes. When the quite option is not
        # set, information about the called script is printed on
        # stdout.
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, quite=False)
        self.assertScriptCalled()
        self.assert_(stdout.startswith('Calling: '))
        self.assert_(stdout.endswith('script\n'))

    def test_observer_not_called_on_outside_change(self):
        # Assert that the script is _not_ called when something
        # outside the observed directory changes.
        observe_dir = os.path.join(self.basedir, 'observe')
        os.mkdir(observe_dir)
        self.callFUT(dir=observe_dir)
        self.assertScriptNotCalled()

    def test_not_existing_script(self):
        # When the script does not exist, an error message is
        # displayed.
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, script='not-existing-script.sh')
        self.assertEqual(
            "OSError: [Errno 2] No such file or directory\n"
            "Parameters were: not-existing-script.sh\n", stdout)

    def test_not_existing_script_not_quite(self):
        # When the script quite option is not set, the file called is
        # displayed, but it is not repeated in the error message.
        self.createObserver()
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, script='not-existing-script.sh', quite=False)
        self.assertEqual(
            "Calling: not-existing-script.sh\n"
            "OSError: [Errno 2] No such file or directory\n", stdout)

    def test_immediate(self):
        # When "immediate" is set the utility is called without a change.
        self.callFUT(immediate=True)
        self.assertScriptCalled()

    def test_immediate_not_quite(self):
        # When "immediate" is set, the "quite" flag is handled, too.
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, quite=False, immediate=True)
        self.assertScriptCalled()
        self.assert_(stdout.startswith('Calling: '))
        self.assert_(stdout.endswith('script\n'))

    def test_immediate_not_existing_script_not_quite(self):
        # When "immediate" is set and the script does not exists, the
        # "quite" flag is handled as usual.
        self.createObserver()
        stdout, ignored = icemac.callonchange.testing.grapStdout(
            self.callFUT, script='not-existing-script.sh', quite=False,
            immediate=True)
        self.assertEqual(
            "Calling: not-existing-script.sh\n"
            "OSError: [Errno 2] No such file or directory\n", stdout)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

import atexit
import fsevents
import optparse
import os.path
import pkg_resources
import subprocess
import sys
import thread


def run_subprocess(quite, params):
    """Run `params` in a subprocess."""
    if not quite:
        print "Calling: %s" % " ".join(params)
    try:
        subprocess.Popen(params)
    except OSError, e:
        # On error it would be nice to have a hint why it failed:
        print "OSError: [Errno %s] %s" % e.args
        if  quite:
            print "Parameters were: %s" % " ".join(params)
        # Exit the observer on error.
        sys.exit(-1)


def run_subprocess_from_thread(quite, params):
    """Run `params` in a subprocess if current thread is not the main one."""
    try:
        run_subprocess(quite, params)
    except SystemExit:
        # Signal the process to exit as we are in a thread here.
        thread.interrupt_main()


def directoryCallbackFactory(quite, *params):
    """Create callback function for directory events."""
    def callback(subpath, mask):
        # Subpath and mask of changes do not matter here. There is
        # currently no way (and no desire) to handle them over to the
        # subprocess.
        run_subprocess_from_thread(quite, params)
    return callback


def fileCallbackFactory(extensions, quite, *params):
    """Create callback function for file events."""
    def callback(event):
        # event.name contains the absolute path to the changed file
        path, ext = os.path.splitext(event.name)
        if ext in extensions:
            # Only run process when extension of changed file is in
            # the list of observed file types
            run_subprocess_from_thread(quite, params)
    return callback


class Observer(object):
    """Observer for path."""

    extensions = []  # only call utility when a file with this ext changed
    quite = False  # when True, do not print any non-error output
    immediate = False  # run immediately after invocation

    def __init__(self, path, params, **options):
        self.path = path
        self.params = params
        self._is_running = False
        for key, value in options.items():
            setattr(self, key, value)

    def start(self):
        if self.immediate:
            run_subprocess(self.quite, self.params)
        if self.extensions:
            # observe explicit file extensions
            callback = fileCallbackFactory(
                self.extensions, self.quite, *self.params)
        else:
            # observe everything in the directory
            callback = directoryCallbackFactory(self.quite, *self.params)
        self.observer = fsevents.Observer()
        self.observer.start()
        file_events = bool(self.extensions)
        self.stream = fsevents.Stream(
            callback, self.path, file_events=file_events)
        self.observer.schedule(self.stream)
        self._is_running = True
        atexit.register(self.stop)

    def stop(self):
        if self._is_running:
            self.observer.unschedule(self.stream)
            self.observer.stop()
            self._is_running = False


def mangle_call_args(args, argv):
    """Combine buildout and sys.argv parameters into one list."""
    call_args = list(tuple(args) + tuple(argv))

    parser = optparse.OptionParser(
        usage="%prog [options] path utility [utility arguments]",
        version=pkg_resources.get_distribution('icemac.callonchange').version)
    parser.description = (
        "%prog invokes <utility> with its <utility arguments> when <path> or "
        "something in it changes.")
    parser.add_option(
        "-e", action="append", metavar="EXTENSION", dest="extensions",
        default=[],
        help="only call utility on changes of a file with this extension "\
             "(option might be used multiple times)")
    parser.add_option(
        "-q", action="store_true", dest="quite", default=False,
        help=("Do not display any output of callonchange. "
              "(Still displays the output of the utility.)"))
    parser.add_option(
        "-i", action="store_true", dest="immediate", default=False,
        help=("Run utility immediately after callonchange has been started."
              "(By default the utility is only run when something changed.)"))
    parser.disable_interspersed_args()

    (options, parsed_args) = parser.parse_args(call_args)

    # At least path and action are required, otherwise we print a
    # helpful message.
    if len(parsed_args) < 2:
        parser.print_help()
        return None, None, None

    # For convinience it is allowed to omit the leading dots in of the
    # specified extension, so they get added here, as they are needed
    # by the observer.
    extensions = []
    for ext in options.extensions:
        if not ext.startswith('.'):
            ext = '.' + ext
        extensions.append(ext)
    options.extensions = extensions

    return parsed_args[0], parsed_args[1:], options.__dict__



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt
from zope.testing import renormalizing
import doctest
import re
import zc.buildout.testing


def setUp(test):
    zc.buildout.testing.buildoutSetUp(test)
    zc.buildout.testing.install_develop("MacFSEvents", test)
    zc.buildout.testing.install_develop("icemac.callonchange", test)


def test_suite():
    return doctest.DocFileSuite(
        'bugfix.txt',
        'recipe.txt',
        setUp=setUp,
        tearDown=zc.buildout.testing.buildoutTearDown,
        optionflags=doctest.ELLIPSIS,
        checker=renormalizing.RENormalizing([
            (re.compile('Not found: /.*/\n'), '')
            ]))



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt


def additional_tests():
    # needed function to find doctests when runing `python setup.py test`
    import test_recipe
    return test_recipe.test_suite()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt
import zc.buildout.easy_install


DEFAULT_CMD_LINE = "-i -e py -e zcml -e pt -e txt src bin/test -cv"
DEFAULT_ARGUMENTS = "'" + "', '".join(DEFAULT_CMD_LINE.split()) + "'"


class Recipe(object):

    def __init__(self, buildout, name, options):
        self.name, self.options = name, options
        self.executable = buildout['buildout']['executable']
        self.bin_directory = buildout['buildout']['bin-directory']
        self.eggs_directory = buildout['buildout']['eggs-directory']
        self.develop_eggs_directory = (
            buildout['buildout']['develop-eggs-directory'])
        links = buildout['buildout'].get('find-links', ())
        if links:
            links = links.split()
        self.links = links
        self.index = buildout['buildout'].get('index')
        self.newest = buildout['buildout'].get('newest') == 'true'
        self.arguments = self.options.get('arguments', DEFAULT_ARGUMENTS)
        name = self.options.get('name', 'callonchange')
        self.scripts = {'callonchange': name}

    def install(self):
        distributions = ['icemac.callonchange']
        ws = zc.buildout.easy_install.install(
            distributions, self.eggs_directory,
            links=self.links,
            index=self.index,
            executable=self.executable,
            path=[self.develop_eggs_directory],
            newest=self.newest)

        return zc.buildout.easy_install.scripts(
            distributions,
            ws,
            self.executable,
            self.bin_directory,
            scripts=self.scripts,
            arguments=self.arguments,
            )

    update = install



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013, 2018 Michael Howitz
# See also LICENSE.txt

import os.path
import setuptools


def read(path):
    """Read a file from a path."""
    with open(os.path.join(*path.split('/'))) as f:
        return f.read()


version = '0.8'

tests_require = [
    'zc.buildout',
    'zope.testing',
]

setuptools.setup(
    name='icemac.callonchange',
    version=version,
    description=(
        "Call a command when a directory or file has changed. (Mac OS only)"),
    long_description=('\n\n'.join([
        read('README.rst'),
        read('CHANGES.rst'),
        read('DOCS.rst'),
        read('TODO.rst'),
        read('HACKING.rst'),
    ])),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Framework :: Buildout :: Recipe',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
    ],
    keywords=(
        'mac os event fs filesystem call change command tdd '
        'test driven development'),
    author='Michael Howitz',
    author_email='icemac@gmx.net',
    url='http://pypi.python.org/icemac.callonchange',
    license='MIT',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['icemac'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'MacFSEvents > 0.2.2, <= 0.2.6',
        'setuptools',
    ],
    extras_require=dict(
        test=tests_require),
    entry_points="""
      [console_scripts]
      callonchange = icemac.callonchange.script:callonchange
      [zc.buildout]
      default = icemac.callonchange.recipe:Recipe
      """,
    tests_require=tests_require,
    test_suite="icemac.callonchange.tests",
)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2011 Michael Howitz
# See also LICENSE.txt

import sys
import time
import icemac.callonchange.observer


def callonchange(*args):
    """Handler for entry point which can handle default and sys.argv parameters."""
    # Combine function arguments which sys.argv arguments:
    path, params, options = icemac.callonchange.observer.mangle_call_args(
        args, sys.argv[1:])
    if path is None:
        # mangle_call_args obviously printed a usage message, so we
        # can quit here.
        return
    observer = icemac.callonchange.observer.Observer(path, params, **options)
    observer.start()
    try:
        try:
            while True:
                # Sleep, waiting for keyboard interrupts does not cost too
                # much resources. The observer runs in a different thread.
                time.sleep(1)
        except KeyboardInterrupt:
            # allow to stop the observer by pressing ^C, this ends the
            # while True loop.
            pass
    finally:
        observer.stop()
